# -*- coding: utf-8 -*-

__title__ = 'bluemoss'
__license__ = 'APACHE'
__version__ = '0.1.30'
__author__ = 'LucaTabone'
__copyright__ = 'Copyright 2023-2023 LucaTabone'

from .classes import *
from .scrape import *
